from enum import Enum


class GraphAPICommonErrorCode(Enum):
    """
    Taken from:
    https://developers.facebook.com/docs/graph-api/guides/error-handling#errorcodes
    """

    API_SESSION = 102
    API_UNKNOWN = 1
    API_SERVICE = 2
    API_METHOD = 3
    API_TOO_MANY_CALLS = 4
    API_PERMISSION_DENIED = 10
    API_USER_TOO_MANY_CALLS = 17
    PAGE_RATE_LIMIT_REACHED = 32
    ACCESS_TOKEN_EXPIRED = 190
    APPLICATION_LIMIT_REACHED = 341
    APPLICATION_BLOCKED_TEMP = 368
    CUSTOM_RATE_LIMIT_REACHED = 613



from __future__ import annotations

import logging
from copy import copy
from types import TracebackType
from typing import Any, Final, Generator, Optional, Type

import backoff
from requests import Response, Session
from requests.exceptions import JSONDecodeError

from .constants import GRAPH_API_URL, GRAPH_API_VERSIONS
from .dataclasses import (
    AppUsageDetails,
    BusinessUseCaseUsageDetails,
    GraphAPIResponse,
    MarketingAPIThrottleInsights,
)
from .error_code import GraphAPICommonErrorCode
from .exceptions import (
    GraphAPIApplicationError,
    GraphAPIError,
    GraphAPIServiceError,
    GraphAPITokenError,
    GraphAPIUsageError,
    InvalidAccessToken,
    InvalidGraphAPIVersion,
)
from .typings import (
    ErrorCodeExceptionMap,
    GraphAPIErrorClassType,
    GraphAPIQueryResult,
    JSONTypeSimple,
)

logger = logging.getLogger(__name__)


class GraphAPIClient:
    """
    A small client built to interact with the Facebook social graph:
    https://developers.facebook.com/docs/graph-api/overview

    This is currently built minimally for distribution in Dash Hudson
    services where JSON-based requests need to be handled.

    The following functionality is currently unsupported when
    comparing to the official facebook-sdk:
        - HMAC authentication
        - batch request handling
        - file uploading
        - generating oauth redirect urls

    For now API access is provisioned through access tokens,
    if you are unfamiliar with how this works see the following:
    https://developers.facebook.com/docs/facebook-login
    """

    DEFAULT_CODE_EXCEPTION_MAP: Final[ErrorCodeExceptionMap] = {
        (GraphAPICommonErrorCode.API_UNKNOWN.value, None): GraphAPIServiceError,
        (GraphAPICommonErrorCode.API_METHOD.value, None): GraphAPIServiceError,
        (GraphAPICommonErrorCode.API_PERMISSION_DENIED.value, None): GraphAPIApplicationError,
        (GraphAPICommonErrorCode.APPLICATION_BLOCKED_TEMP.value, None): GraphAPIApplicationError,
        (GraphAPICommonErrorCode.API_SESSION.value, None): GraphAPITokenError,
        (GraphAPICommonErrorCode.ACCESS_TOKEN_EXPIRED.value, None): GraphAPITokenError,
        (GraphAPICommonErrorCode.APPLICATION_LIMIT_REACHED.value, None): GraphAPIUsageError,
        (GraphAPICommonErrorCode.API_TOO_MANY_CALLS.value, None): GraphAPIUsageError,
        (GraphAPICommonErrorCode.PAGE_RATE_LIMIT_REACHED.value, None): GraphAPIUsageError,
        (GraphAPICommonErrorCode.CUSTOM_RATE_LIMIT_REACHED.value, None): GraphAPIUsageError,
    }

    def __init__(
        self,
        access_token: str,
        version: str,
        global_timeout: Optional[int] = None,
        params_to_mask: Optional[list[str]] = None,
        retry_params: Optional[dict] = None,
        disable_logger: Optional[bool] = False,
        code_exception_map: Optional[ErrorCodeExceptionMap] = None,
        loose_match_errors: Optional[bool] = False,
    ) -> None:
        """
        Initialize the API client
        :param access_token: An access token provisioned through Facebook login
        :param version: The Graph API version to use (ex: 12.0)
        :param global_timeout: A global request timeout to set
        :param params_to_mask: A list of query parameter names to mask when formatting
            exception messages
        :param disable_logger Disables exception logging if truthy
        :param retry_config Params for https://github.com/litl/backoff#backoffon_exception
        :param code_exception_map: A an error code -> exception map / configuration
        """
        if not access_token or not isinstance(access_token, str):
            raise InvalidAccessToken
        version = (
            version[1:] if isinstance(version, str) and version.lower().startswith('v') else version
        )
        if version not in GRAPH_API_VERSIONS:
            raise InvalidGraphAPIVersion(version)

        self.version = f'v{version}'
        self.global_timeout = global_timeout
        self.params_to_mask = params_to_mask
        self.disable_logger = disable_logger
        # Defaulting to max_tries=0 disables retrying by default
        self.retry_params = retry_params or {'exception': tuple(), 'max_tries': 0}

        self.code_exception_map = self.DEFAULT_CODE_EXCEPTION_MAP
        if code_exception_map:
            self.code_exception_map = {
                **self.DEFAULT_CODE_EXCEPTION_MAP,
                **code_exception_map,
            }

        self._access_token = access_token
        self._session = Session()
        self._session.params = {'access_token': self._access_token}

        self._loose_match_errors = loose_match_errors

    def get(
        self,
        path: str,
        params: Optional[dict] = None,
        timeout: Optional[int] = None,
        retry_params: Optional[dict] = None,
        **kwargs: Any,
    ) -> GraphAPIResponse:
        """
        :param path: A path pointing to an edge or node
            (ex: /<page_id>/conversations)
        Performs a GET request to the Graph API
        :param params: Query parameters to be included with the request
        :param timeout: A custom timeout for the request (seconds)
        :param retry_params: Retry params override
        :return: An instance of GraphAPIResponse
        """
        return self._do_request(
            method='GET',
            path=path,
            params=params,
            timeout=timeout,
            retry_params=retry_params,
            **kwargs,
        )

    def get_all_pages(
        self,
        path: str,
        params: Optional[dict] = None,
        timeout: Optional[int] = None,
        retry_params: Optional[dict] = None,
        **kwargs: Any,
    ) -> Generator[GraphAPIResponse, None, None]:
        """
        :param path: A path pointing to an edge or node
            (ex: /<page_id>/conversations)
        Performs a GET request to the Graph API
        :param params: Query parameters to be included with the request
        :param timeout: A custom timeout for the request (seconds)
        :param retry_params: Retry params override
        :return: An iterator containing paginated instances of GraphAPIResponse
        """
        params = copy(params) if params else {}
        params['after'] = None
        while True:
            res = self._do_request(
                method='GET',
                path=path,
                params=params,
                timeout=timeout,
                retry_params=retry_params,
                **kwargs,
            )
            yield res
            if not res.after_cursor or not res.next_page_url:
                break
            params['after'] = res.after_cursor

    def get_all_pages_from_next_url(
        self,
        next_url: str,
        timeout: Optional[int] = None,
        retry_params: Optional[dict] = None,
        **kwargs: Any,
    ) -> Generator[GraphAPIResponse, None, None]:
        _next_url = next_url
        while True:
            res = self._do_request(
                method='GET',
                full_url=_next_url,
                timeout=timeout,
                retry_params=retry_params,
                **kwargs,
            )
            yield res
            if not res.next_page_url:
                break
            _next_url = res.next_page_url

    def post(
        self,
        path: str,
        data: Any,
        params: Optional[Any] = None,
        timeout: Optional[int] = None,
        retry_params: Optional[dict] = None,
        **kwargs: Any,
    ) -> GraphAPIResponse:
        """
        Performs a POST request to the Graph API
        :param path: A path pointing to an edge or node
            (ex: /<page_id>/conversations | /<page_id>)
        :param data: The request body to be included
        :param params: Query parameters to be included with the request
        :param timeout: A custom timeout for the request (seconds)
        :param retry_params: Retry params override
        :return: An instance of GraphAPIResponse
        """
        return self._do_request(
            method='POST',
            path=path,
            params=params,
            data=data,
            timeout=timeout,
            retry_params=retry_params,
            **kwargs,
        )

    def delete(
        self,
        path: str,
        params: Optional[dict] = None,
        timeout: Optional[int] = None,
        retry_params: Optional[dict] = None,
        **kwargs: Any,
    ) -> GraphAPIResponse:
        """
        Performs a DELETE request to the Graph API
        :param path: A path pointing to a node
            (ex: /<video_id>)
        :param params: Query parameters to be included with the request
        :param timeout: A custom timeout for the request (seconds)
        :param retry_params: Retry params override
        :return: An instance of GraphAPIResponse
        """
        return self._do_request(
            method='DELETE',
            path=path,
            params=params,
            timeout=timeout,
            retry_params=retry_params,
            **kwargs,
        )

    def _do_request(
        self,
        method: str,
        path: str = '',
        full_url: str = '',
        params: Optional[Any] = None,
        data: Optional[Any] = None,
        timeout: Optional[int] = None,
        retry_params: Optional[dict] = None,
        **kwargs: Any,
    ) -> GraphAPIResponse:
        """
        Handle Graph API requests. Raise if error body detected and lets ambiguous network
        errors propagate.
        :param method: The HTTP request method
        :param path: A path pointing to an edge or node (ex: /<page_id>/conversations)
        :param params: Query parameters to be included with the request
        :param data: The request body to be included
        :param timeout: A custom timeout for the request (seconds)
        :param retry_params: Retry params override
        :return: An instance of GraphAPIResponse
        """

        @backoff.on_exception(backoff.expo, **(retry_params or self.retry_params))
        def _retry_parameterizer() -> GraphAPIResponse:
            if not path and not full_url:
                raise ValueError('either path or full_url must be specified')

            response = self._session.request(
                method=method,
                url=f'{GRAPH_API_URL}/{self.version}/{path}' if path else full_url,
                params=params,
                data=data,
                timeout=timeout or self.global_timeout,
            )
            result, paging = self._parse_response_body_or_raise(response)

            return GraphAPIResponse(
                app_usage_details=AppUsageDetails.from_header(response),
                business_use_case_usage_details=BusinessUseCaseUsageDetails.from_header(response),
                marketing_api_throttle_insights=MarketingAPIThrottleInsights.from_header(response),
                data=result,
                paging=paging,
            )

        return _retry_parameterizer()

    def _parse_response_body_or_raise(
        self, response: Response
    ) -> tuple[GraphAPIQueryResult, Optional[JSONTypeSimple]]:
        """
        Parse Graph API response body and raise if error details present
        :param response: A response from the Graph API
        :return: Parsed request body and optional paging params
        """
        try:
            response_body = response.json()
        except JSONDecodeError:
            logger.exception(f'Failed to parse response body: {response.text}')
            raise GraphAPIError(response, {'message': 'Failed to parse response body'})
        error_details = response_body.get('error')
        if error_details:
            # Raise a specific exception if a code mapping is set, custom exceptions take priority
            exc_type = self._get_exc_type(error_details)
            # Log & raise default GraphAPIError if no mapping was found
            exc = exc_type(
                response=response,
                error_details=error_details,
                params_to_mask=self.params_to_mask,
            )
            if not self.disable_logger:
                logger.error(str(exc))
            raise exc
        # If 'data' is present, it means the result is a list of graph nodes and may have
        # paging params as well
        if 'data' in response_body:
            return response_body['data'], response_body.get('paging')
        # If not, the response body is a single graph node without paging params
        return response_body, None

    def _get_exc_type(self, error_details: dict[str, Any]) -> GraphAPIErrorClassType:
        code_key: tuple[Any, Any] = (error_details.get('code'), error_details.get('error_subcode'))
        # Raise a specific exception if a code mapping is set, custom exceptions take priority
        if exc_type := self.code_exception_map.get(code_key):
            return exc_type
        # If no mapping was found, try to match loosely
        if self._loose_match_errors and code_key[1]:
            exc_type = self.code_exception_map.get((code_key[0], None))
        return exc_type or GraphAPIError

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        self._session.close()



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional

from requests import Response

from .helpers import deserialize_json_header
from .typings import GraphAPIQueryResult, JSONTypeSimple


@dataclass(frozen=True)
class BaseUsageDetails:
    call_count: int
    total_time: int
    total_cputime: int


@dataclass(frozen=True)
class AppUsageDetails(BaseUsageDetails):
    """
    Encapsulates stats from X-App-Usage header:
    https://developers.facebook.com/docs/graph-api/overview/rate-limiting#headers
    """

    @classmethod
    def from_header(cls, res: Response) -> AppUsageDetails:
        app_usage_dict = deserialize_json_header(res=res, header_name='X-App-Usage')
        return cls(
            call_count=app_usage_dict.get('call_count', 0),
            total_time=app_usage_dict.get('total_time', 0),
            total_cputime=app_usage_dict.get('total_cputime', 0),
        )


@dataclass(frozen=True)
class BusinessUseCaseUsageDetails(BaseUsageDetails):
    """
    Encapsulates stats from X-Business-Use-Case-Usage header:
    https://developers.facebook.com/docs/graph-api/overview/rate-limiting#headers-2
    """

    type: str | None
    estimated_time_to_regain_access: int

    @classmethod
    def from_header(cls, res: Response) -> dict[str, BusinessUseCaseUsageDetails]:
        buc_usage_dict = deserialize_json_header(res=res, header_name='X-Business-Use-Case-Usage')
        return_dict = {}
        for id_, usage_dicts in buc_usage_dict.items():
            if not (usage_dict := next(iter(usage_dicts), None)):
                continue
            time_to_regain_access = usage_dict.get('estimated_time_to_regain_access') or 0
            return_dict[id_] = cls(
                type=usage_dict.get('type'),
                estimated_time_to_regain_access=time_to_regain_access,
                call_count=usage_dict.get('call_count', 0),
                total_time=usage_dict.get('total_time', 0),
                total_cputime=usage_dict.get('total_cputime', 0),
            )
        return return_dict


@dataclass(frozen=True)
class MarketingAPIThrottleInsights:
    """
    Encapsulates stats from X-Fb-Ads-Insights-Throttle header:
    https://developers.facebook.com/docs/marketing-api/insights/best-practices/#insightscallload
    """

    app_id_util_pct: float
    acc_id_util_pct: float
    ads_api_access_tier: str

    @classmethod
    def from_header(cls, res: Response) -> MarketingAPIThrottleInsights:
        throttle_insights_dict = deserialize_json_header(
            res=res, header_name='X-Fb-Ads-Insights-Throttle'
        )
        return cls(
            app_id_util_pct=throttle_insights_dict.get('app_id_util_pct', 0.0),
            acc_id_util_pct=throttle_insights_dict.get('acc_id_util_pct', 0.0),
            ads_api_access_tier=throttle_insights_dict.get('ads_api_access_tier', ''),
        )


@dataclass
class GraphAPIResponse:
    """
    Encapsulates a Graph API response payload with parsed app usage headers
    """

    app_usage_details: AppUsageDetails
    business_use_case_usage_details: dict[str, BusinessUseCaseUsageDetails]
    marketing_api_throttle_insights: MarketingAPIThrottleInsights
    data: GraphAPIQueryResult
    paging: Optional[JSONTypeSimple] = None

    @property
    def is_empty(self) -> bool:
        return not self.data

    @property
    def is_list(self) -> bool:
        return isinstance(self.data, list)

    @property
    def is_dict(self) -> bool:
        return isinstance(self.data, dict)

    @property
    def before_cursor(self) -> Optional[str]:
        return self.cursors.get('before')

    @property
    def after_cursor(self) -> Optional[str]:
        return self.cursors.get('after')

    @property
    def next_page_url(self) -> Optional[str]:
        return self.paging.get('next') if self.paging else None

    @property
    def cursors(self) -> JSONTypeSimple:
        return self.paging.get('cursors', {}) if self.paging else {}



from .client import GraphAPIClient
from .dataclasses import AppUsageDetails, BusinessUseCaseUsageDetails, GraphAPIResponse
from .error_code import GraphAPICommonErrorCode
from .exceptions import (
    GraphAPIApplicationError,
    GraphAPIError,
    GraphAPIServiceError,
    GraphAPITokenError,
    GraphAPIUsageError,
    InvalidAccessToken,
    InvalidGraphAPIVersion,
)
from .helpers import FieldConfig, build_field_config_list, format_fields_str
from .typings import ErrorCodeExceptionMap, GraphAPIErrorClassType

__all__ = [
    'GraphAPIClient',
    'AppUsageDetails',
    'BusinessUseCaseUsageDetails',
    'GraphAPIResponse',
    'GraphAPICommonErrorCode',
    'GraphAPIApplicationError',
    'GraphAPIError',
    'GraphAPIServiceError',
    'GraphAPITokenError',
    'GraphAPIUsageError',
    'InvalidAccessToken',
    'InvalidGraphAPIVersion',
    'FieldConfig',
    'build_field_config_list',
    'format_fields_str',
    'ErrorCodeExceptionMap',
    'GraphAPIErrorClassType',
]



import json
from typing import Any, Final, Iterable

import requests
from typing_extensions import NotRequired, TypedDict

# TODO: Recursive type hints not yet available in mypy,
#  should change sub_fields hint to Optional[List[FieldConfig]] when possible:
#  https://github.com/python/mypy/issues/731
FieldConfig = TypedDict(
    'FieldConfig',
    {
        'field_name': str,
        'limit': NotRequired[int],
        'after': NotRequired[str],
        'date_preset': NotRequired[str],
        'sub_fields': NotRequired[list[Any]],
    },
)


ACCEPTED_SUB_PARAM_NAMES: Final = (
    'limit',
    'after',
    'date_preset',
    'breakdowns',
    'metric',
)


def build_field_config_list(field_literals: Iterable[str]) -> list[FieldConfig]:
    """
    Helper function to convert iterables containing field literals into FieldConfig objects
        for use with format_fields_str
    :param field_literals: An iterable containing a Graph API field literal (i.e. ['foo']).
    :return: A list containing respective FieldConfig objects (i.e. [{'field_name': 'foo'}]).
    """
    return [{'field_name': f} for f in field_literals]


def format_fields_str(fields_config: list[FieldConfig]) -> str:
    """
    Helper function to support field expansion / limiting functionality:
    https://developers.facebook.com/docs/graph-api/field-expansion
    :param fields_config: The field configurations you want contained in string format
    :return: A formatted fields string
    """
    fields_str = ''
    for config in fields_config:
        # Prep the base sub string for a field w/ params (i.e. foo.limit(5).after(BAR))
        fields_str += f'{config["field_name"]}'
        for param_name in ACCEPTED_SUB_PARAM_NAMES:
            fields_str += _format_sub_parameter_str(config, param_name)
        # If sub-fields exist, prep the nested sub-field(s) strings
        sub_fields = config.get('sub_fields')
        if sub_fields:
            fields_str += f'{{{format_fields_str(sub_fields)}}}'

        # If not the last item in a config list, add a comma separator
        if config != fields_config[-1]:
            fields_str += ','
    return fields_str


def _format_sub_parameter_str(config: FieldConfig, param_name: str) -> str:
    """
    Used to help format sub-param strings for a specific field:
    Ex: https://developers.facebook.com/docs/graph-api/field-expansion#limiting-results
    :param config: An instance of FieldConfig
    :param param_name: The parameter name to pluck from config (i.e. limit or after)
    :return: A formatted paging sub string (i.e. .limit(5))
    """
    if not (param := config.get(param_name)):
        return ''
    if isinstance(param, (list, set)):
        param = ','.join(param)
    return f'.{param_name}({param})'


def deserialize_json_header(res: requests.Response, header_name: str) -> dict:
    header_val = res.headers.get(header_name)
    return json.loads(header_val) if header_val else {}



# Just here to save some boiler, recursive JSON structure is difficult in mypy:
# https://github.com/python/typing/issues/182
from typing import Any, Optional, Type, Union

from .exceptions import GraphAPIError

JSONTypeSimple = dict[str, Any]
GraphAPIErrorClassType = Type[GraphAPIError]
ErrorCodeExceptionMap = dict[tuple[int, Optional[int]], GraphAPIErrorClassType]
GraphAPIQueryResult = Union[JSONTypeSimple, list[JSONTypeSimple]]



from typing import Any, Final, Optional, Set, Union
from urllib.parse import parse_qs, urlsplit

from requests import Response

from .constants import GRAPH_API_VERSIONS


class InvalidAccessToken(Exception):
    """
    Raised if invalid access token supplied to Client
    """

    def __init__(self) -> None:
        self.message = 'Invalid access token: must be of type str'

    def __str__(self) -> str:  # pragma: no cover
        return self.message


class InvalidGraphAPIVersion(Exception):
    """
    Raised if invalid version supplied to Client
    """

    def __init__(self, version: Any) -> None:
        self.message = (
            f'Invalid Graph API version: {version}. '
            f'Available versions: {", ".join(GRAPH_API_VERSIONS)}'
        )

    def __str__(self) -> str:  # pragma: no cover
        return self.message


class GraphAPIError(Exception):
    """
    Encapsulate Graph API & HTTP error data. Raised when error response
    detected from Graph API
    """

    SENSITIVE_FIELD_MASK: Final = '*****'
    DEFAULT_PARAMS_TO_MASK: Final = ['access_token']
    DETAIL_NOT_FOUND_PLACEHOLDER: Final = '<not_found>'

    def __init__(
        self,
        response: Response,
        error_details: dict[str, Union[str, int]],
        params_to_mask: Optional[list[str]] = None,
    ) -> None:
        """
        Scrapes Graph API error body and HTTP response
        :param response: An instance of requests.Response encapsulating the graph API call
        :param error_details: Parsed error details / body from Graph API,
            this is done ahead of time in Client
        :param params_to_mask: A list of additional query params to mask in error messages
        """
        super().__init__(response, error_details, params_to_mask)
        self.status_code = response.status_code
        self.reason = response.reason
        self.fb_error_msg = error_details.get('message', GraphAPIError.DETAIL_NOT_FOUND_PLACEHOLDER)
        self.type = error_details.get('type', GraphAPIError.DETAIL_NOT_FOUND_PLACEHOLDER)
        self.code = error_details.get('code', GraphAPIError.DETAIL_NOT_FOUND_PLACEHOLDER)
        self.subcode = error_details.get(
            'error_subcode', GraphAPIError.DETAIL_NOT_FOUND_PLACEHOLDER
        )
        self.fbtrace_id = error_details.get(
            'fbtrace_id', GraphAPIError.DETAIL_NOT_FOUND_PLACEHOLDER
        )

        split_url = urlsplit(response.url)
        self.path = split_url.path
        # Ensure default param masks are always set
        self.sanitized_query_params = GraphAPIError._sanitize_query_params(
            split_url.query,
            set(GraphAPIError.DEFAULT_PARAMS_TO_MASK + (params_to_mask or [])),
        )

        self.message = (
            f'{self.fb_error_msg}\n'
            '[Debug Info]\n'
            f'Type: {self.type}\n'
            f'Code: {self.code}\n'
            f'Subcode: {self.subcode}\n'
            f'Trace ID: {self.fbtrace_id}\n'
            '[HTTP Info]\n'
            f'Status Code: {self.status_code} | {self.reason}\n'
            f'Path: {self.path}\n'
            f'Query Params: {self.sanitized_query_params}'
        )

    def __str__(self) -> str:  # pragma: no cover
        return self.message

    @staticmethod
    def _sanitize_query_params(query: str, params_to_mask: Set[str]) -> dict[str, list[str]]:
        """
        Process query params present on request that was sent and mask sensitive fields
        :param response: Response to pull query params from
        :param params_to_mask: The parameters that should have their values masked
        :return: Transformed query params dict where sensitive params are masked
        """
        parsed_params = parse_qs(query)
        for param in params_to_mask:
            if param in parsed_params:
                parsed_params[param] = [
                    GraphAPIError.SENSITIVE_FIELD_MASK for _ in parsed_params[param]
                ]
        return parsed_params


class GraphAPITokenError(GraphAPIError):
    """
    Raised if a token issue occurs
    """

    pass


class GraphAPIServiceError(GraphAPIError):
    """
    Raised if an unknown API error occurs
    """

    pass


class GraphAPIApplicationError(GraphAPIError):
    """
    Raised if there's an issue with the developer application associated with a token
    """

    pass


class GraphAPIUsageError(GraphAPIError):
    """
    Raised if the developer application is rate limited or throttled
    """

    pass


class GraphAPIUserError(GraphAPIError):
    """
    Raised if there's an issue with the user account associated with a token
    """

    pass
